package ComplexNumbers.ComplexNumbers;


import lists.LinkedListCharacters.Structures.LinkedListCharacters;

import static java.lang.Math.*;
import static lists.LinkedListCharacters.LinkedListCharactersFunctions.LinkedListCharactersFunctions.*;
import static mathmath.math.math.*;
import static mathmath.math.math.mathSinh;
import static numbers.NumberToString.NumberToString.CreateStringDecimalFromNumber;

public class ComplexNumbers {
    public static ComplexNumber CreateComplexNumber(double re, double im) {
        ComplexNumber z;

        z = new ComplexNumber();
        z.re = re;
        z.im = im;

        return z;
    }

    public static PolarComplexNumber CreatePolarComplexNumber(double r, double phi) {
        PolarComplexNumber p;

        p = new PolarComplexNumber();
        p.r = r;
        p.phi = phi;

        return p;
    }

    public static void Add(ComplexNumber z1, ComplexNumber z2){
        double a, b, c, d;

        a = z1.re;
        b = z1.im;
        c = z2.re;
        d = z2.im;

        z1.re = a + c;
        z1.im = b + d;
    }

    public static ComplexNumber AddToNew(ComplexNumber z1, ComplexNumber z2){
        ComplexNumber x;
        double a, b, c, d;

        a = z1.re;
        b = z1.im;
        c = z2.re;
        d = z2.im;

        x = new ComplexNumber();

        x.re = a + c;
        x.im = b + d;

        return x;
    }

    public static void Sub(ComplexNumber z1, ComplexNumber z2){
        double a, b, c, d;

        a = z1.re;
        b = z1.im;
        c = z2.re;
        d = z2.im;

        z1.re = a - c;
        z1.im = b - d;
    }

    public static ComplexNumber SubToNew(ComplexNumber z1, ComplexNumber z2){
        ComplexNumber x;
        double a, b, c, d;

        a = z1.re;
        b = z1.im;
        c = z2.re;
        d = z2.im;

        x = new ComplexNumber();

        x.re = a - c;
        x.im = b - d;

        return x;
    }

    public static void Mul(ComplexNumber z1, ComplexNumber z2){
        double a, b, c, d;

        a = z1.re;
        b = z1.im;
        c = z2.re;
        d = z2.im;

        z1.re = a*c - b*d;
        z1.im = b*c + a*d;
    }

    public static ComplexNumber MulToNew(ComplexNumber z1, ComplexNumber z2){
        ComplexNumber x;
        double a, b, c, d;

        a = z1.re;
        b = z1.im;
        c = z2.re;
        d = z2.im;

        x = new ComplexNumber();

        x.re = a*c - b*d;
        x.im = b*c + a*d;

        return x;
    }

    public static void Div(ComplexNumber z1, ComplexNumber z2){
        double a, b, c, d;

        a = z1.re;
        b = z1.im;
        c = z2.re;
        d = z2.im;

        z1.re = (a*c + b*d)/(pow(c, 2d) + pow(d, 2d));
        z1.im = (b*c - a*d)/(pow(c, 2d) + pow(d, 2d));
    }

    public static ComplexNumber DivToNew(ComplexNumber z1, ComplexNumber z2){
        ComplexNumber x;
        double a, b, c, d;

        a = z1.re;
        b = z1.im;
        c = z2.re;
        d = z2.im;

        x = new ComplexNumber();

        x.re = (a*c + b*d)/(pow(c, 2d) + pow(d, 2d));
        x.im = (b*c - a*d)/(pow(c, 2d) + pow(d, 2d));

        return x;
    }

    public static void Conjugate(ComplexNumber z){
        z.im = -z.im;
    }

    public static ComplexNumber ConjugateToNew(ComplexNumber z){
        ComplexNumber x;

        x = new ComplexNumber();

        x.re = z.re;
        x.im = -z.im;

        return x;
    }

    public static double Abs(ComplexNumber z){
        double x;

        x = sqrt(pow(z.re, 2d) + pow(z.im, 2d));

        return x;
    }

    public static double Arg(ComplexNumber z){
        double x;

        x = mathAtan2(z.im, z.re);

        return x;
    }

    public static PolarComplexNumber CreatePolarFromComplexNumber(ComplexNumber z){
        PolarComplexNumber x;

        x = new PolarComplexNumber();

        x.r = Abs(z);
        x.phi = Arg(z);

        return x;
    }

    public static ComplexNumber CreateComplexFromPolar(PolarComplexNumber p){
        ComplexNumber z;

        z = new ComplexNumber();

        z.re = p.r * cos(p.phi);
        z.im = p.r * sin(p.phi);

        return z;
    }

    public static double Re(ComplexNumber z){
        return z.re;
    }

    public static double Im(ComplexNumber z){
        return z.im;
    }

    public static void AddPolar(PolarComplexNumber p1, PolarComplexNumber p2){
        PolarComplexNumber x;
        ComplexNumber z1, z2;

        z1 = CreateComplexFromPolar(p1);
        z2 = CreateComplexFromPolar(p2);

        Add(z1, z2);

        x = CreatePolarFromComplexNumber(z1);

        p1.r = x.r;
        p1.phi = x.phi;

        delete(z1);
        delete(z2);
        delete(x);
    }

    public static PolarComplexNumber AddPolarToNew(PolarComplexNumber p1, PolarComplexNumber p2){
        PolarComplexNumber x;
        ComplexNumber z1, z2;

        z1 = CreateComplexFromPolar(p1);
        z2 = CreateComplexFromPolar(p2);

        Add(z1, z2);

        x = CreatePolarFromComplexNumber(z1);

        delete(z1);
        delete(z2);

        return x;
    }

    public static void SubPolar(PolarComplexNumber p1, PolarComplexNumber p2){
        PolarComplexNumber x;
        ComplexNumber z1, z2;

        z1 = CreateComplexFromPolar(p1);
        z2 = CreateComplexFromPolar(p2);

        Sub(z1, z2);

        x = CreatePolarFromComplexNumber(z1);

        p1.r = x.r;
        p1.phi = x.phi;

        delete(z1);
        delete(z2);
        delete(x);
    }

    public static PolarComplexNumber SubPolarToNew(PolarComplexNumber p1, PolarComplexNumber p2){
        PolarComplexNumber x;
        ComplexNumber z1, z2;

        z1 = CreateComplexFromPolar(p1);
        z2 = CreateComplexFromPolar(p2);

        Sub(z1, z2);

        x = CreatePolarFromComplexNumber(z1);

        delete(z1);
        delete(z2);

        return x;
    }

    public static void MulPolar(PolarComplexNumber p1, PolarComplexNumber p2){
        double r1, r2, phi1, phi2;

        r1 = p1.r;
        r2 = p2.r;
        phi1 = p1.phi;
        phi2 = p2.phi;

        p1.r = r1 * r2;
        p1.phi = phi1 + phi2;
    }

    public static PolarComplexNumber MulPolarToNew(PolarComplexNumber p1, PolarComplexNumber p2){
        PolarComplexNumber x;
        double r1, r2, phi1, phi2;

        r1 = p1.r;
        r2 = p2.r;
        phi1 = p1.phi;
        phi2 = p2.phi;

        x = new PolarComplexNumber();

        x.r = r1 * r2;
        x.phi = phi1 + phi2;

        return x;
    }

    public static void DivPolar(PolarComplexNumber p1, PolarComplexNumber p2){
        double r1, r2, phi1, phi2;

        r1 = p1.r;
        r2 = p2.r;
        phi1 = p1.phi;
        phi2 = p2.phi;

        p1.r = r1 / r2;
        p1.phi = phi1 - phi2;
    }

    public static PolarComplexNumber DivPolarToNew(PolarComplexNumber p1, PolarComplexNumber p2){
        PolarComplexNumber x;
        double r1, r2, phi1, phi2;

        r1 = p1.r;
        r2 = p2.r;
        phi1 = p1.phi;
        phi2 = p2.phi;

        x = new PolarComplexNumber();

        x.r = r1 / r2;
        x.phi = phi1 - phi2;

        return x;
    }

    public static void SquareRoot(ComplexNumber z){
        double a, b, m;

        a = z.re;
        b = z.im;

        m = sqrt(pow(a, 2d) + pow(b, 2d));

        z.re = sqrt((m + a)/2d);
        z.im = mathSign(b) * sqrt((m - a)/2d);
    }

    public static void PowerPolar(PolarComplexNumber p, double n){
        // De Moivre's Theorem
        p.r = pow(p.r, n);
        p.phi = p.phi * n;
    }

    public static ComplexNumber PowerToNew(ComplexNumber z, double n){
        PolarComplexNumber p;
        ComplexNumber zp;

        p = CreatePolarFromComplexNumber(z);
        PowerPolar(p, n);
        zp = CreateComplexFromPolar(p);

        delete(p);

        return zp;
    }

    public static void Power(ComplexNumber z, double n){
        ComplexNumber zp;

        zp = PowerToNew(z, n);
        z.re = zp.re;
        z.im = zp.im;

        delete(zp);
    }

    public static void Negate(ComplexNumber z){
        z.re = mathNegate(z.re);
        z.im = mathNegate(z.im);
    }

    public static void delete(Object object) {
        // Java has garbage collection.
    }

    public static void AssignComplexByValues(ComplexNumber s, double re, double im) {
        s.re = re;
        s.im = im;
    }

    public static void AssignComplex(ComplexNumber a, ComplexNumber b) {
        a.re = b.re;
        a.im = b.im;
    }

    public static boolean EpsilonCompareComplex(ComplexNumber a, ComplexNumber b, double epsilon){
        return mathEpsilonCompare(a.re, b.re, epsilon) && mathEpsilonCompare(a.im, b.im, epsilon);
    }

    public static void ExpComplex(ComplexNumber x){
        double re, im;

        re = exp(x.re)*cos(x.im);
        im = exp(x.re)*sin(x.im);
        x.re = re;
        x.im = im;
    }

    public static void SineComplex(ComplexNumber x){
        double re, im;

        re = sin(x.re)*mathCosh(x.im);
        im = cos(x.re)*mathSinh(x.im);
        x.re = re;
        x.im = im;
    }

    public static void CosineComplex(ComplexNumber x){
        double re, im;

        re = cos(x.re)*mathCosh(x.im);
        im = sin(x.re)*mathSinh(x.im);
        x.re = re;
        x.im = im;
    }

    public static char[] ComplexToString(ComplexNumber a) {
        char [] str, number;
        LinkedListCharacters ll;
        double i;

        ll = CreateLinkedListCharacter();

        number = CreateStringDecimalFromNumber(a.re);

        for(i = 0d; i < number.length; i = i + 1d) {
            LinkedListAddCharacter(ll, number[(int)i]);
        }

        delete(number);

        if(a.im < 0d){
            LinkedListAddCharacter(ll, '-');
            number = CreateStringDecimalFromNumber(-a.im);
        }else{
            LinkedListAddCharacter(ll, '+');
            number = CreateStringDecimalFromNumber(a.im);
        }

        for(i = 0d; i < number.length; i = i + 1d) {
            LinkedListAddCharacter(ll, number[(int)i]);
        }

        delete(number);

        LinkedListAddCharacter(ll, 'i');

        str = LinkedListCharactersToArray(ll);
        FreeLinkedListCharacter(ll);

        return str;
    }

    /*public static boolean CreateComplexNumberFromString(String s, ComplexNumber a) {
        return false;
    }*/
}















